#--------------------------------------------------------------------------
#
#   HXFS 2019 - Flexagon
#
#--------------------------------------------------------------------------

default_face_colors = [
    (1, 0, 0),
    (0, 1, 0),
    (0, 0, 1),
    (1, 1, 0),
    (1, 0.5, 0),
    (0.5, 0, 1),
    (1, 1, 1),
    (0, 0, 0),
]

class Triangle(object):

    def __init__(self, pat, state, depth = 0, front = 0, back = 0):
        self.pat = pat
        self.depth = depth
        self.state = state
        self.faces = [front, back]
    
    def __repr__(self):
        return "Triangle(pat=%s,depth=%s,state=%s)" % (self.pat, self.depth, self.state)
    
    def dump(self):
        print "%s,%s: #%s" % (self.pat, self.depth, self.index), self.state, self.faces

    def permute_state(self, i, j, k):
        s = self.state
        self.state = (s[i], s[j], s[k])
    
    def top_face_index(self):
        s = self.state
        if (s[1] - s[0]) % 3 == 1:
            return 0
        else:
            return 1
    
    def get_top_and_bottom_faces(self):
        f = self.faces
        i = self.top_face_index()
        return f[i], f[1 - i]
    
    def set_top_and_bottom_faces(self, top, bottom):
        f = self.faces
        i = self.top_face_index()
        f[i] = top
        f[1 - i] = bottom        

#--------------------------------------------------------------------------

class Flexagon(object):

    def __init__(self):
        self.face_colors = default_face_colors[:]
        self.pats = [[] for i in xrange(6)]
        t = []
        for i in xrange(0, 6, 2):
            t.append(Triangle(pat = i, depth = 0, state = (0, 2, 1), back = 0, front = 2))
            t.append(Triangle(pat = i, depth = 1, state = (2, 1, 0), back = 2, front = 1))
            t.append(Triangle(pat = i+1, depth = 0, state = (0, 1, 2), front = 0, back = 1))
        index_triangles(t)
        self.triangles = t
        self.update_flexability()
        print "New flexagon:"
        self.dump()
    
    def dump(self):
        #pats = self.pats_in_depth_order()
        pats = self.pats_in_linkage_order()
        for k, pat in enumerate(pats):
            for t in pat:
                print "%s,%s: #%s" % (k, t.depth, t.index), t.state, t.faces
        print "flexability", self.corner_flexability
        print
    
    def pats_in_linkage_order(self):
        p = [[] for k in xrange(6)]
        k0 = 0
        triangles = self.triangles
        while triangles[k0].pat == triangles[-1].pat:
            k0 += 1
        n = len(triangles)
        for i in xrange(n):
            k = (k0 + i) % n
            t = triangles[k]
            p[t.pat].append(t)
        return p
    
    def pats_in_depth_order(self):
        return convert_pats_to_depth_order(self.pats_in_linkage_order())
    
    def pat_in_depth_order(self, k):
        pat = [t for t in self.triangles if t.pat == k]
        pat.sort(key = triangle_depth)
        return pat
    
    def update_flexability(self):
        pats = self.pats_in_linkage_order()
        cf = [0 for k in xrange(6)]
        for k, pat in enumerate(pats):
            d1 = pat[0].depth
            d2 = pat[-1].depth
            if d1 < d2:
                cf[k] |= 1
                cf[(k - 1) % 6] |= 2
            elif d1 > d2:
                cf[k] |= 2
                cf[(k - 1) % 6] |= 1
        self.corner_flexability = cf
    
    def check_flex(self, i, b):
        f = self.corner_flexability
        for j in xrange(i, i+6, 2):
            if not (f[j] & b):
                print ("Flex", "Antiflex")[b-1], "of corner", j, "is not possible"
                return False
        return True

#     def flex(self, k):
#         i0 = k & 1
#         if not self.check_flex(i0, 1):
#             return
#         print "Flexing corner", k
#         plo = self.pats_in_linkage_order()
#         pdo = convert_pats_to_depth_order(plo)
#         for i in xrange(i0, i0 + 6, 2):
#             j = (i + 1) % 6
#             print "Flexing pats", i, "and", j
#             pi = pdo[i]
#             pj = pdo[j]
#             print "Pat", i, "in depth order:"
#             dump_pat(pi)
#             print "Pat", j, "in depth order:"
#             dump_pat(pj)
#             ptop = []
#             pmid = []
#             pbot = []
#             din = plo[i][0].depth
#             print "Coming in at depth", din
#             for t in pi:
#                 if t.depth <= din:
#                     print "Flipping #%s into top" % t.index
#                     t.permute_state(0, 2, 1)
#                     ptop.insert(0, t)
#                     t.dump()
#                 else:
#                     print "Rotating #%s into middle" % t.index
#                     t.permute_state(2, 0, 1)
#                     pmid.append(t)
#                     t.dump()
#             for t in pj:
#                 print "Flipping #%s into bottom" % t.index
#                 t.permute_state(2, 1, 0)
#                 pbot.insert(0, t)
#                 t.dump()
#             form_pat(ptop, i)
#             form_pat(pmid + pbot , j)
#         self.update_flexability()
#         print
#         self.dump()

    def flex(self, k):
        i0 = k & 1
        if not self.check_flex(i0, 1):
            return
        print "Flexing corner", k
        plo = self.pats_in_linkage_order()
        for i in xrange(i0, i0 + 6, 2):
            j = (i + 1) % 6
            print "Flexing pats", i, "and", j
            ploi = plo[i]
            ploj = plo[j]

            din = ploi[0].depth
            dout = ploi[-1].depth
            print "Coming in at depth", din
            print "Going out at depth", dout
            
            ptop = []
            pmid = []
            pisplit = (ptop, pmid)
            for t in ploi:
                if t.depth <= din:
                    print "#%s is above incoming depth" % t.index
                    h = 0
                elif t.depth >= dout:
                    print "#%s is below outgoing depth" % t.index
                    h = 1
                else:
                    print "#%s is %s split" % (("above", "below")[h], t.index)
                pisplit[h].append(t)
            
            ptop.sort(key = reverse_triangle_depth)
            pmid.sort(key = triangle_depth)
            pbot = sorted(ploj, key = reverse_triangle_depth)
            
            for t in ptop:
                print "Flipping #%s into top" % t.index
                t.permute_state(0, 2, 1)
            for t in pmid:
                print "Rotating #%s into middle" % t.index
                t.permute_state(2, 0, 1)
            for t in pbot:
                print "Flipping #%s into bottom" % t.index
                t.permute_state(2, 1, 0)
            
            form_pat(ptop, i)
            form_pat(pmid + pbot , j)
        self.update_flexability()
        print
        self.dump()

#     def antiflex(self, k):
#         i0 = k & 1
#         if not self.check_flex(i0, 2):
#             return
#         print "Antiflexing corner", k
#         plo = self.pats_in_linkage_order()
#         pdo = convert_pats_to_depth_order(plo)
#         for i in xrange(i0, i0 + 6, 2):
#             j = (i + 1) % 6
#             pi = pdo[i]
#             pj = pdo[j]
#             ptop = []
#             pmid = []
#             pbot = []
#             din = plo[j][-1].depth
#             for t in pi:
#                 t.permute_state(0, 2, 1)
#                 ptop.insert(0, t)
#             for t in pj:
#                 if t.depth >= din:
#                     t.permute_state(2, 1, 0)
#                     pbot.insert(0, t)
#                 else:
#                     t.permute_state(1, 2, 0)
#                     pmid.append(t)
#             form_pat(ptop + pmid, i)
#             form_pat(pbot, j)
#         self.update_flexability()
#         self.dump()
    
    def antiflex(self, k):
        i0 = k & 1
        if not self.check_flex(i0, 2):
            return
        print "Antiflexing corner", k
        plo = self.pats_in_linkage_order()
        for i in xrange(i0, i0 + 6, 2):
            j = (i + 1) % 6
            ploi = plo[i]
            ploj = plo[j]

            din = ploj[-1].depth
            dout = ploj[0].depth
            
            pmid = []
            pbot = []
            pjsplit = (pmid, pbot)
            for t in ploj:
                if t.depth >= din:
                    h = 1
                elif t.depth <= dout:
                    h = 0
                pjsplit[h].append(t)
            
            ptop = sorted(ploi, key = reverse_triangle_depth)
            pmid.sort(key = triangle_depth)
            pbot.sort(key = reverse_triangle_depth)
            
            for t in ptop:
                t.permute_state(0, 2, 1)
            for t in pmid:
                t.permute_state(1, 2, 0)
            for t in pbot:
                t.permute_state(2, 1, 0)
            
            form_pat(ptop + pmid, i)
            form_pat(pbot , j)
        self.update_flexability()
        print

    def find_unused_color(self):
        n = len(self.face_colors)
        candidates = set(xrange(n))
        for t in self.triangles:
            for c in t.faces:
                candidates.discard(c)
        candidates.add(n - 1)
        return min(candidates)
    
    def split_top_of_pat(self, k, new_face_color):
        pat = self.pat_in_depth_order(k)
        t = pat[0]
        t1 = Triangle(pat = k, state = permute_sides(t.state, (0, 2, 1)))
        t2 = Triangle(pat = k, state = permute_sides(t.state, (2, 1, 0)))
        if t.top_face_index() == 1:
            print "Exchanging new triangles"
            t1, t2 = t2, t1
        top, bottom = t.get_top_and_bottom_faces()
        t1.set_top_and_bottom_faces(top, new_face_color)
        t2.set_top_and_bottom_faces(new_face_color, bottom)
        triangles = self.triangles
        i = t.index
        triangles[i] = t1
        triangles.insert(i + 1, t2)
        index_triangles(triangles)
        pat[0] = t1
        pat.insert(1, t2)
        form_pat(pat, k)
        self.update_flexability()

#--------------------------------------------------------------------------

def index_triangles(tlist):
    for i, t in enumerate(tlist):
        t.index = i

def triangle_depth(t):
    return t.depth

def reverse_triangle_depth(t):
    return -t.depth

def permute_sides(state, perm):
    return tuple(perm[i] for i in state)

def form_pat(p, i):
    for d, t in enumerate(p):
        t.pat = i
        t.depth = d

def convert_pats_to_depth_order(pats):
    return [sorted(p, key = triangle_depth) for p in pats]

def dump_pat(pat):
    for t in pat:
        t.dump()

#--------------------------------------------------------------------------

if __name__ == "__main__":
    f = Flexagon()
    f.dump()
    f.flex(1)
    f.flex(0)
    f.dump()
    f.flex(0)
    f.antiflex(0)
    f.dump()
    f.antiflex(0)
